#ifndef BWASEARCH_H
#define BWASEARCH_H


#include <algorithm>
#include <vector>

#include <misc/utils.h>


namespace CSA
{

/*
  This uses the RANGES part of external module interface.
*/
  
//--------------------------------------------------------------------------

struct WeightedMatchInfo
{
  const static usint MATCH = 0x1;
  const static usint MISMATCH = 0x2;
  const static usint INSERTION = 0x4;
  const static usint DELETION = 0x8;

  const static usint MISMATCH_PENALTY = 3;
  const static usint GAP_OPEN = 11;
  const static usint GAP_EXTEND = 4;

  usint     score;
  usint     errors;
  usint     type;
  usint     matched;
  bool      rc;
  pair_type range;

  WeightedMatchInfo() :
    score(0), errors(0), type(MATCH),
    matched(0), rc(false), range(0, 0)
  {
  }

  WeightedMatchInfo(bool reverse_complement, pair_type rng) :
    score(0), errors(0), type(MATCH),
    matched(0), rc(reverse_complement), range(rng)
  {
  }

  WeightedMatchInfo(usint s, usint e, usint t, usint m, bool r, pair_type rng) :
    score(s), errors(e), type(t),
    matched(m), rc(r), range(rng)
  {
  }

  WeightedMatchInfo match(pair_type rng) const
  {
    return WeightedMatchInfo(this->score, this->errors, MATCH, this->matched + 1, this->rc, rng);
  }

  WeightedMatchInfo mismatch(pair_type rng) const
  {
    return WeightedMatchInfo(this->score + MISMATCH_PENALTY, this->errors + 1, MISMATCH,
                             this->matched + 1, this->rc, rng);
  }

  WeightedMatchInfo insertion() const
  {
    usint penalty = ((this->type == INSERTION) ? GAP_EXTEND : GAP_OPEN);
    return WeightedMatchInfo(this->score + penalty, this->errors + 1, INSERTION,
                             this->matched + 1, this->rc, this->range);
  }

  WeightedMatchInfo deletion(pair_type rng) const
  {
    usint penalty = ((this->type == DELETION) ? GAP_EXTEND : GAP_OPEN);
    return WeightedMatchInfo(this->score + penalty, this->errors + 1, DELETION,
                             this->matched, this->rc, rng);
  }

  bool operator<(const WeightedMatchInfo& another) const
  {
    return (this->score > another.score); // Our priority queue is a maximum heap.
  }
};

//--------------------------------------------------------------------------

struct MatchInfo
{
  usint      matched;
  bool       rc;
  pair_type  range;

  MatchInfo() :
    matched(0), rc(false), range(0, 0)
  {
  }

  MatchInfo(usint _matched, bool reverse_complement, pair_type _range) :
    matched(_matched), rc(reverse_complement), range(_range)
  {
  }

  MatchInfo(const WeightedMatchInfo& info) :
    matched(info.matched), rc(info.rc), range(info.range)
  {
  }

  MatchInfo match(pair_type rng) const
  {
    return MatchInfo(this->matched + 1, this->rc, rng);
  }

  MatchInfo mismatch(pair_type rng) const
  {
    return MatchInfo(this->matched + 1, this->rc, rng);
  }

  MatchInfo insertion() const
  {
    return MatchInfo(this->matched + 1, this->rc, this->range);
  }

  MatchInfo deletion(pair_type rng) const
  {
    return MatchInfo(this->matched, this->rc, rng);
  }

  bool operator<(const MatchInfo& another) const
  {
    if(this->matched < another.matched) { return true; }
    if(this->matched > another.matched) { return false; }

    if(!this->rc && another.rc) { return true; }
    if(this->rc && !another.rc) { return false; }

    return (this->range < another.range);
  }

  bool operator==(const MatchInfo& another) const
  {
    return ((this->matched == another.matched) &&
            (this->rc == another.rc) &&
            (this->range == another.range));
  }
};

//--------------------------------------------------------------------------

template<class Index>
class BWASearch
{
  public:

    const static usint INITIAL_STEP = 8;  // Should be >= log(n) / 4.

    BWASearch(const Index& _index) :
      index(_index), ALPHABET("ACGT"), COMPLEMENT("TGCA")
    {
    }

    pair_type find(const std::string& pattern, bool reverse_complement, usint skip = 0) const
    {
      std::string pat = pattern.substr(skip);
      if(pat.length() == 0) { return this->index.getSARange(); }

      char c = (reverse_complement ? this->complement(pat[0]) : pat[pat.length() - 1]);
      pair_type range = this->index.getCharRange(c);
      if(isEmpty(range)) { return range; }

      MatchInfo info(1, reverse_complement, range);
      this->backwardSearch(pat, info);
      this->index.convertToSARange(info.range);
      
      return info.range;
    }

    /*
      Approximate search with at most k mismatches/errors for the pattern and its
      reverse complement. Returns only the best matches.
      User must free the returned vector.
      We do not match any character to 'Z' in this search.

      The algorithm is similar to the one in:

      H. Li, R. Durbin: Fast and accurate short read alignment with
      Burrows-Wheeler transform. Bioinformatics, 2009.
    */
    std::vector<pair_type>* find(const std::string& pattern, usint k, bool allow_indels) const
    {
      std::vector<pair_type>* results = new std::vector<pair_type>;
      std::vector<MatchInfo>* current_matches = new std::vector<MatchInfo>;
      std::vector<MatchInfo>* next_matches = new std::vector<MatchInfo>;

      pair_type range = this->index.getBWTRange();
      usint* fw_bounds = this->errorBounds(pattern, k, false);
      if(fw_bounds[0] <= k)
      {
        current_matches->push_back(MatchInfo(0, false, range));
      }
      usint* rc_bounds = this->errorBounds(pattern, k, true);
      if(rc_bounds[0] <= k)
      {
        current_matches->push_back(MatchInfo(0, true, range));
      }

      for(usint errors = 0; errors <= k; errors++)
      {
        if(allow_indels) { removeDuplicates(current_matches, false); }

        while(!(current_matches->empty()))
        {
          MatchInfo info = current_matches->back(); current_matches->pop_back();
          if(info.matched >= pattern.length())
          {
            results->push_back(info.range);
            k = errors;
            continue;
          }

          if(errors == k)
          {
            this->backwardSearch(pattern, info);
            if(!isEmpty(info.range)) { results->push_back(info.range); }
            continue;
          }

          char c = (info.rc ? this->complement(pattern[info.matched]) :
                              pattern[pattern.length() - info.matched - 1]);
          usint* bounds = (info.rc ? rc_bounds : fw_bounds);
          const std::string& alphabet = (info.rc ? this->COMPLEMENT : this->ALPHABET);

          pair_type rng = this->index.LF(info.range, c);
          if(!isEmpty(rng)) { current_matches->push_back(info.match(rng)); }
          if(bounds[info.matched + 1] < k - errors)
          {
            for(usint i = 0; i < ALPHABET_SIZE; i++)
            {
              if(alphabet[i] == c) { continue; }
              rng = this->index.LF(info.range, alphabet[i]);
              if(isEmpty(rng)) { continue; }
              next_matches->push_back(info.mismatch(rng));

              if(allow_indels && bounds[info.matched] < k - errors)
              {
                next_matches->push_back(info.deletion(rng));
              }
            }
            if(allow_indels)
            {
              next_matches->push_back(info.insertion());
            }
          }
        }

        if(next_matches->empty()) { break; }
        delete current_matches; current_matches = next_matches;
        if(errors < k) { next_matches = new std::vector<MatchInfo>; } else { next_matches = 0; }
      }

      delete current_matches; current_matches = 0;
      delete next_matches; next_matches = 0;
      delete[] fw_bounds; fw_bounds = 0;
      delete[] rc_bounds; rc_bounds = 0;

      removeDuplicates(results, false);
      this->index.convertToSARange(*results);
      return results;
    }

    // Same as above, but with different penalties for different edit operations.
    // Additionally, we do not allow extending gaps.
    std::vector<pair_type>* weightedFind(const std::string& pattern, usint k, bool allow_indels) const
    {
      std::vector<pair_type>* results = new std::vector<pair_type>;
      std::vector<WeightedMatchInfo> heap;

      pair_type range = this->index.getBWTRange();
      usint* fw_bounds = this->errorBounds(pattern, k, false);
      if(fw_bounds[0] <= k)
      {
        heap.push_back(WeightedMatchInfo(false, range));
        push_heap(heap.begin(), heap.end());
      }
      usint* rc_bounds = this->errorBounds(pattern, k, true);
      if(rc_bounds[0] <= k)
      {
        heap.push_back(WeightedMatchInfo(true, range));
        push_heap(heap.begin(), heap.end());
      }

      usint best_match = WORD_MAX;
      while(!(heap.empty()))
      {
        pop_heap(heap.begin(), heap.end());
        WeightedMatchInfo info = heap.back(); heap.pop_back();
        if(info.score > best_match) { break; }
        if(info.matched >= pattern.length())
        {
          results->push_back(info.range);
          best_match = std::min(best_match, info.score);
          continue;
        }

        if(info.errors == k)
        {
          MatchInfo mi(info);
          this->backwardSearch(pattern, mi);
          if(!isEmpty(mi.range))
          {
            results->push_back(mi.range);
            best_match = std::min(best_match, info.score);
          }
          continue;
        }

        char c = (info.rc ? this->complement(pattern[info.matched]) :
                            pattern[pattern.length() - info.matched - 1]);
        usint* bounds = (info.rc ? rc_bounds : fw_bounds);
        const std::string& alphabet = (info.rc ? this->COMPLEMENT : this->ALPHABET);

        pair_type rng = this->index.LF(info.range, c);
        if(!isEmpty(rng))
        {
          heap.push_back(info.match(rng));
          push_heap(heap.begin(), heap.end());
        }
        if(bounds[info.matched + 1] < k - info.errors &&
           info.score + WeightedMatchInfo::MISMATCH_PENALTY <= best_match)
        {
          for(usint i = 0; i < ALPHABET_SIZE; i++)
          {
            if(alphabet[i] == c) { continue; }
            rng = this->index.LF(info.range, alphabet[i]);
            if(isEmpty(rng)) { continue; }
            heap.push_back(info.mismatch(rng));
            push_heap(heap.begin(), heap.end());

            if(allow_indels && bounds[info.matched] < k - info.errors &&
               (info.type & (WeightedMatchInfo::MATCH | WeightedMatchInfo::DELETION)) != 0)
            {
              heap.push_back(info.deletion(rng));
              push_heap(heap.begin(), heap.end());
            }
          }
          if(allow_indels &&
             (info.type & (WeightedMatchInfo::MATCH | WeightedMatchInfo::INSERTION)) != 0)
          {
            heap.push_back(info.insertion());
            push_heap(heap.begin(), heap.end());
          }
        }
      }

      delete[] fw_bounds; fw_bounds = 0;
      delete[] rc_bounds; rc_bounds = 0;
      removeDuplicates(results, false);
      this->index.convertToSARange(*results);
      return results;
    }

    bool handleOccurrences(pair_type range, bool locate, usint& n)
    {
      if(isEmpty(range)) { return false; }
      
      if(locate)
      {
        std::vector<usint>* occurrences = this->index.locateRange(range);
        if(occurrences != 0)
        {
          n += occurrences->size();
          delete occurrences;
        }
      }

      return true;
    }

    bool handleOccurrences(std::vector<pair_type>* ranges, bool locate, usint& n)
    {
      if(ranges->empty()) { return false; }

      if(locate)
      {
        std::vector<usint>* occurrences = this->index.locateRanges(*ranges);
        if(occurrences != 0)
        {
          n += occurrences->size();
          delete occurrences;
        }
      }

      return true;
    }

  private:
    const Index& index;

    const std::string ALPHABET;
    const std::string COMPLEMENT;
    static const usint ALPHABET_SIZE = 4;

    char complement(char c) const
    {
      size_t temp = this->COMPLEMENT.find(c);
      if(temp != std::string::npos) { c = this->ALPHABET[temp]; }
      return c;
    }

    void backwardSearch(const std::string& pattern, MatchInfo& info) const
    {
      sint pos = (info.rc ? info.matched : pattern.length() - info.matched - 1);
      sint adj = (info.rc ? 1 : -1);

      for(; info.matched < pattern.length(); info.matched++, pos += adj)
      {
        char c = (info.rc ? this->complement(pattern[pos]) : pattern[pos]);
        info.range = this->index.LF(info.range, c);
        if(isEmpty(info.range)) { return; }
      }
    }

    usint* errorBounds(const std::string& pattern, usint k, bool complement) const
    {
      // bounds[i]: lower bound for errors after matching 'i' characters.
      usint* bounds = new usint[pattern.length() + 1];

      usint errors = 0;
      usint start = 0, limit = pattern.length() - 1;
      while(start <= limit)
      {
        // Search for the lowest 'low' such that there is no match for 'pattern[start, low]'.
        usint low = start, high = limit + 1;
        usint diff = INITIAL_STEP;

        if(errors < k)
        {
          while(low < high && low <= limit)
          {
            usint mid = std::min(low + 2 * diff, low + (high - low) / 2);
            diff = mid - low;
            MatchInfo info(0, complement, this->index.getBWTRange());
            usint temp = (complement ? pattern.length() - mid - 1 : start);
            this->backwardSearch(pattern.substr(temp, mid + 1 - start), info);
            if(isEmpty(info.range))
            {
              high = mid;
            }
            else
            {
              low = mid + 1;
            }
          }

          for(usint i = start; i < low; i++) { bounds[i] = errors; }
          errors++; start = low + 1;
          if(low <= limit) { bounds[low] = errors; }
        }
        else  // We cannot afford any more errors.
        {
          MatchInfo info(0, complement, this->index.getBWTRange());
          usint temp = (complement ? 0 : start);
          this->backwardSearch(pattern.substr(temp, pattern.length() - start), info);
          if(isEmpty(info.range)) { errors++; }
          for(usint i = start; i <= limit; i++) { bounds[i] = errors; }
          break;
        }
      }

      std::reverse(bounds, bounds + pattern.length());
      bounds[pattern.length()] = 0;
      return bounds;
    }

    // These are not allowed.
    BWASearch();
    BWASearch(const BWASearch&);
    BWASearch& operator = (const BWASearch&);
};

//--------------------------------------------------------------------------

};  // namespace CSA


#endif  // BWASEARCH_H