#include <algorithm>
#include <fstream>
#include <iostream>

#include <misc/utils.h>

#include "gcsa.h"


namespace CSA
{

const std::string GCSA::ALPHABET = "ACGTZ";

//--------------------------------------------------------------------------

GCSA::GCSA(const std::string& base_name) :
  ok(false), support_locate(false),
  outgoing(0),
  sampled_positions(0), samples(0)
{
  for(usint i = 0; i < ALPHABET_SIZE; i++) { this->array[i] = 0; }

  std::string index_name = base_name + GCSA_EXTENSION;
  std::ifstream input(index_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "Error opening input file (" << index_name << ")!" << std::endl;
    return;
  }

  input.read((char*)(this->counts), ALPHABET_SIZE * sizeof(usint));
  this->fillArrayC();

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i] = new DeltaVector(input);
  }
  this->outgoing = new RLEVector(input);

  this->sampled_positions = new DeltaVector(input);
  usint sample_bits = 0;
  input.read((char*)&sample_bits, sizeof(usint));
  if(sample_bits > 0)
  {
    this->samples = new ReadBuffer(input, this->sampled_positions->getNumberOfItems(), sample_bits);
    this->support_locate = true;
  }

  this->ok = true;
  input.close();
}

GCSA::GCSA(PathGraph& graph, bool print) :
  ok(false), support_locate(false),
  outgoing(0),
  sampled_positions(0), samples(0)
{
  if(graph.status != PathGraph::sorted) { return; }

  DeltaEncoder* array_encoders[ALPHABET_SIZE];
  RLEEncoder outedges(OUTGOING_BLOCK_SIZE);

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i] = 0;
    this->counts[i] = 0;
    array_encoders[i] = new DeltaEncoder(ARRAY_BLOCK_SIZE);
  }
  if(print) { std::cout << "Generating edges... "; std::cout.flush(); }
  if(!graph.generateEdges()) { return; }
  if(print) { std::cout << graph.edge_count << " edges." << std::endl; }


  usint offset = 0, edge_offset = 0;
  for(std::vector<PathNode>::iterator node = graph.nodes.begin(); node != graph.nodes.end(); ++node)
  {
    // Write BWT.
    for(usint i = 0; i < ALPHABET_SIZE; i++)
    {
      if(node->hasPred(i))
      {
        this->counts[i]++;
        array_encoders[i]->setBit(offset);
      }
    }
    offset++;

    // Write M
    outedges.addBit(edge_offset);
    edge_offset += std::max((usint)1, (*node).outdegree());
  }
  this->fillArrayC();

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i] = new DeltaVector(*(array_encoders[i]), offset);
  }
  outedges.flush();
  this->outgoing = new RLEVector(outedges, edge_offset);

  // Sample the graph for locate().
  if(print) { std::cout << "Sampling the graph... "; std::cout.flush();  }
  std::vector<pair_type>* sample_pairs = graph.getSamples(SAMPLE_RATE);
  usint max_sample = 0;
  for(usint i = 0; i < sample_pairs->size(); i++)
  {
    if(sample_pairs->at(i).second > max_sample) { max_sample = sample_pairs->at(i).second; }
  }

  DeltaEncoder sample_encoder(SAMPLE_BLOCK_SIZE);
  WriteBuffer sample_values(sample_pairs->size(), length(max_sample));
  for(usint i = 0; i < sample_pairs->size(); i++)
  {
    sample_encoder.addBit(sample_pairs->at(i).first);
    sample_values.writeItem(sample_pairs->at(i).second);
  }
  sample_encoder.flush();

  this->sampled_positions = new DeltaVector(sample_encoder, offset);
  this->samples = sample_values.getReadBuffer();
  this->support_locate = true;
  delete sample_pairs; sample_pairs = 0;
  if(print)
  {
    std::cout << this->sampled_positions->getNumberOfItems() << " samples." << std::endl;
    std::cout << std::endl;
  }


  if(print)
  {
    std::cout << "Nodes:           " << graph.node_count << std::endl;
    std::cout << "Edges:           " << graph.edge_count << std::endl;
    std::cout << "BWT length:      " << offset << std::endl;
    std::cout << "Samples:         " << this->sampled_positions->getNumberOfItems() << std::endl;
    std::cout << std::endl;

    this->reportSize(true);
  }
  this->ok = true;
}

GCSA::~GCSA()
{
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    delete this->array[i]; this->array[i] = 0;
  }
  delete this->outgoing; this->outgoing = 0;
  delete this->sampled_positions; this->sampled_positions = 0;
  delete this->samples; this->samples = 0;
}

void
GCSA::writeTo(const std::string& base_name) const
{
  std::string index_name = base_name + GCSA_EXTENSION;
  std::ofstream output(index_name.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "Error opening output file (" << index_name << ")!" << std::endl;
    return;
  }

  output.write((char*)(this->counts), ALPHABET_SIZE * sizeof(usint));
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i]->writeTo(output);
  }
  this->outgoing->writeTo(output);

  this->sampled_positions->writeTo(output);
  usint sample_bits = (this->support_locate ? this->samples->getItemSize() : 0);
  output.write((char*)&sample_bits, sizeof(usint));
  if(this->support_locate) { this->samples->writeBuffer(output); }

  output.close();
}

bool
GCSA::isOk() const
{
  return this->ok;
}

//--------------------------------------------------------------------------

usint
GCSA::reportSize(bool print) const
{
  usint array_size = 0;
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    array_size += this->array[i]->reportSize();
  }

  usint outedges = this->outgoing->reportSize();

  usint sa_samples = this->sampled_positions->reportSize();
  if(this->support_locate) { sa_samples += this->samples->reportSize(); }

  usint bytes = sizeof(*this) + array_size + outedges + sa_samples;
  if(print)
  {
    std::cout << "Array:           " << (array_size / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << "Outgoing edges:  " << (outedges / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << "Samples:         " << (sa_samples / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << "Total size:      " << (bytes / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << std::endl;
  }

  return bytes;
}

void
GCSA::printCounts() const
{
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    std::cout << "counts[" << ALPHABET[i] << "] = " << this->counts[i] << std::endl;
  }
  std::cout << std::endl;
}

usint
GCSA::getSize() const
{
  return this->sampled_positions->getSize();
}

//--------------------------------------------------------------------------

pair_type
GCSA::find(const std::string& pattern) const
{
  if(pattern.length() == 0) { return this->getSARange(); }

  sint pos = pattern.length() - 1;
  pair_type range = this->getCharRange(pattern[pos]);
  if(isEmpty(range)) { return range; }

  for(pos--; pos >= 0; pos--)
  {
    range = this->LF(range, pattern[pos]);
    if(isEmpty(range)) { return range; }
  }
  return range;
}

//--------------------------------------------------------------------------

std::vector<usint>*
GCSA::locate(pair_type range) const
{
  if(!(this->support_locate)) { return 0; }
  std::vector<usint>* vec = new std::vector<usint>;

  this->locate(range, *vec);
  removeDuplicates(vec, false);

  return vec;
}

std::vector<usint>*
GCSA::locate(std::vector<pair_type>& ranges) const
{
  if(!(this->support_locate)) { return 0; }
  std::vector<usint>* vec = new std::vector<usint>;

  for(std::vector<pair_type>::iterator iter = ranges.begin(); iter != ranges.end(); ++iter)
  {
    this->locate(*iter, *vec);
  }
  removeDuplicates(vec, false);

  return vec;
}

//--------------------------------------------------------------------------

pair_type
GCSA::getSARange() const
{
  return pair_type(0, this->getSize() - 1);
}

pair_type
GCSA::getBWTRange() const
{
  return pair_type(0, this->getSize() - 1);
}

pair_type
GCSA::getCharRange(usint c) const
{
  if(c == 0) { return pair_type(0, 0); }

  size_t temp = ALPHABET.find(c);
  if(temp == std::string::npos || this->counts[temp] == 0) { return EMPTY_PAIR; }

  return this->convertToNodeRange(pair_type(this->cumulative[temp], this->cumulative[temp + 1] - 1));
}

void
GCSA::convertToBWTRange(pair_type& sa_range) const
{
}

void
GCSA::convertToSARange(pair_type& bwt_range) const
{
}

void
GCSA::convertToSARange(std::vector<pair_type>& bwt_ranges) const
{
}

pair_type
GCSA::LF(pair_type range, usint c) const
{
  size_t temp = ALPHABET.find(c);
  if(temp == std::string::npos || this->counts[temp] == 0) { return EMPTY_PAIR; }

  // Follow edges backward using BWT.
  DeltaVector::Iterator array_iter(*(this->array[temp]));
  range.first = this->cumulative[temp] + array_iter.rank(range.first, true) - 1;
  range.second = this->cumulative[temp] + array_iter.rank(range.second) - 1;
  if(isEmpty(range)) { return EMPTY_PAIR; }

  return this->convertToNodeRange(range);
}

std::vector<usint>*
GCSA::locateRange(pair_type range) const
{
  return this->locate(range);
}

std::vector<usint>*
GCSA::locateRanges(std::vector<pair_type>& ranges) const
{
  return this->locate(ranges);
}

//--------------------------------------------------------------------------

bool
GCSA::exists(usint c) const
{
  size_t temp = ALPHABET.find(c);
  if(temp == std::string::npos) { return false; }

  return (this->counts[temp] > 0);
}

usint
GCSA::getCharacter(usint index) const
{
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    if(this->cumulative[i + 1] > index) { return i; }
  }

  return ALPHABET_SIZE;
}

pair_type
GCSA::convertToNodeRange(pair_type edge_range) const
{
  RLEVector::Iterator outgoing_iter(*(this->outgoing));
  edge_range.first = outgoing_iter.rank(edge_range.first) - 1;
  edge_range.second = outgoing_iter.rank(edge_range.second) - 1;
  return edge_range;
}

usint
GCSA::Psi(usint index) const
{
  if(index == 0) { return this->getSize() - 1; } // Final node.

  RLEVector::Iterator outgoing_iter(*(this->outgoing));
  index = outgoing_iter.select(index);
  usint c = this->getCharacter(index);

  // Find the corresponding incoming edge using BWT.
  DeltaVector::Iterator array_iter(*(this->array[c]));
  index = array_iter.select(index - this->cumulative[c]);

  return index;
}

//--------------------------------------------------------------------------

void
GCSA::locate(pair_type range, std::vector<usint>& vec) const
{
  range.second = std::min(range.second, this->getSize() - 1);
  if(isEmpty(range)) { return; }

  DeltaVector::Iterator sample_iter(*(this->sampled_positions));
  for(usint pos = range.first; pos <= range.second; pos++)
  {
    usint temp = pos, steps = 0;
    while(!(sample_iter.isSet(temp)))
    {
      temp = this->Psi(temp); steps++;
    }
    vec.push_back(this->samples->readItem(sample_iter.rank(temp) - 1) - steps);
  }
}

//--------------------------------------------------------------------------

void
GCSA::fillArrayC()
{
  this->cumulative[0] = 1;

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->cumulative[i + 1] = this->cumulative[i] + this->counts[i];
  }
}

//--------------------------------------------------------------------------

} // namespace CSA
