#include <algorithm>
#include <cstdlib>
#include <sstream>
#include <queue>
#include <vector>
#include <stack>

#include "graph.h"
#include "gcsa.h"

#include <misc/utils.h>
#include <bits/bitbuffer.h>


namespace CSA
{

//--------------------------------------------------------------------------

Graph::Graph(std::vector<GraphNode>& node_vector, std::vector<GraphEdge>& edge_vector) :
  nodes(0), edges(0),
  node_count(node_vector.size()), edge_count(edge_vector.size()),
  ok(false)
{
  GraphNode* node = this->nodes = new GraphNode[this->node_count];
  for(std::vector<GraphNode>::iterator iter = node_vector.begin();
      iter != node_vector.end(); ++iter, ++node)
  {
    *node = *iter;
  }

  GraphEdge* edge = this->edges = new GraphEdge[this->edge_count];
  for(std::vector<GraphEdge>::iterator iter = edge_vector.begin();
      iter != edge_vector.end(); ++iter, ++edge)
  {
    *edge = *iter;
  }

  this->ok = true;
}

Graph::Graph(const std::string& base_name) :
  nodes(0), edges(0),
  node_count(0), edge_count(0),
  ok(false)
{
  std::string automaton_name = base_name + AUTOMATON_EXTENSION;
  std::ifstream infile(automaton_name.c_str(), std::ios_base::binary);
  if(!infile)
  {
    std::cerr << "Graph: Cannot open input file (" << automaton_name << ")!" << std::endl;
    return;
  }

  infile.read((char*)&(this->node_count), sizeof(this->node_count));
  infile.read((char*)&(this->edge_count), sizeof(this->edge_count));
  this->nodes = new GraphNode[this->node_count];
  infile.read((char*)(this->nodes), this->node_count * sizeof(GraphNode));
  this->edges = new GraphEdge[this->edge_count];
  infile.read((char*)(this->edges), this->edge_count * sizeof(GraphEdge));

  infile.close();
  this->sortEdges(true);  // FIXME unnecessary, as the edges are already sorted?
  this->ok = true;
}

Graph::~Graph()
{
  delete[] this->nodes; this->nodes = 0;
  delete[] this->edges; this->edges = 0;
}

void
Graph::write(const std::string& base_name)
{       
  if(!this->ok) { return; }

  std::string automaton_name = base_name + AUTOMATON_EXTENSION;
  std::ofstream outfile(automaton_name.c_str(), std::ios_base::binary);
  if(!outfile)
  {
    std::cerr << "Graph: Cannot open output file (" << automaton_name << ")!" << std::endl;
    return;
  }

  outfile.write((char*)&(this->node_count), sizeof(this->node_count));
  outfile.write((char*)&(this->edge_count), sizeof(this->edge_count));
  outfile.write((char*)(this->nodes), this->node_count * sizeof(GraphNode));
  outfile.write((char*)(this->edges), this->edge_count * sizeof(GraphEdge));

  outfile.close();
}

void Graph::printInfo()
{
  std::cout << "Generation 0 (" << this->node_count << " nodes, " << this->edge_count << " edges)" << std::endl;
}

struct GraphEdgeFromComparator
{
  bool operator() (const GraphEdge& a, const GraphEdge& b) const
  {
    return (a.from < b.from);
  }
} ge_from_comparator;

struct GraphEdgeToComparator
{
  bool operator() (const GraphEdge& a, const GraphEdge& b) const
  {
    return (a.to < b.to);
  }
} ge_to_comparator;

void
Graph::sortEdges(bool from)
{
  if(from) { parallelSort(this->edges, this->edges + this->edge_count, ge_from_comparator); }
  else
  {
    parallelSort(this->edges, this->edges + this->edge_count, ge_to_comparator);
    this->ok = false;
  }
}

pair_type
Graph::getEdges(uint node, bool from)
{
  pair_type range;
  uint      temp;

  // Lower bound for the range.
  sint low = 0, high = (usint)(this->edge_count) - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    temp = (from ? this->edges[mid].from : this->edges[mid].to);
    if(node == temp)     { high = mid; }
    else if(node < temp) { high = mid - 1; }
    else                 { low = mid + 1; }
  }
  temp = (from ? this->edges[low].from : this->edges[low].to);
  if(node == temp) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = (usint)(this->edge_count) - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    temp = (from ? this->edges[mid].from : this->edges[mid].to);
    if(node == temp) { low = mid; }
    else             { high = mid - 1; }
  }
  range.second = high;

  return range;
}

//--------------------------------------------------------------------------

PathGraph::PathGraph(Graph& _parent) :
  node_count(0), edge_count(0), ranks(0),
  parent(_parent), generation(1),
  status(error)
{
  if(!this->parent.ok) { return; }

  // One PathNode per edge and one extra for the end marker.
  this->node_count = this->parent.edge_count + 1;
  this->nodes.reserve(this->node_count);
  for(GraphEdge* edge = this->parent.edges;
      edge < this->parent.edges + this->parent.edge_count; ++edge)
  {
    this->createPathNode(edge->from, edge->to);
  }
  this->createPathNode(this->parent.node_count - 1, this->parent.node_count - 1);

  this->status = ok;
  this->sort();
}

PathGraph::PathGraph(PathGraph& previous) :
  node_count(0), edge_count(0), ranks(0),
  parent(previous.parent), generation(previous.generation + 1),
  status(error)
{
  if(previous.status != ok || !this->parent.ok) { return; }

  this->nodes.reserve(previous.node_count + previous.node_count / 4);  // Overestimate?
  previous.sortByFrom();

  for(nvector::iterator left = previous.nodes.begin(); left != previous.nodes.end(); ++left)
  {
    if((*left).isSorted()) { this->nodes.push_back(*left); continue; }

    pair_type ge_range = this->parent.getEdges((*left).to, true);
    for(usint ge = ge_range.first; ge <= ge_range.second; ge++)
    {
      pair_type pn_range = previous.getNodesFrom(this->parent.edges[ge].to);
      for(usint pn = pn_range.first; pn <= pn_range.second; pn++)
      {
        this->createPathNode(*left, previous.nodes[pn]);
      }
    }
  }
  this->node_count = this->nodes.size();

  this->status = ok;
  this->sort();
}

PathGraph::~PathGraph()
{
  this->status = error;
}

void
PathGraph::printInfo()
{
  std::cout << "Generation " << this->generation << " (" << this->node_count << " nodes, " << this->edge_count << " edges, " << this->ranks << " ranks)" << std::endl;
}

bool
PathGraph::generateEdges()
{
  if(this->status != sorted) { return false; }

  this->edges.reserve(this->node_count + this->node_count / 4); // Overestimate?
  this->parent.sortEdges(false);

  // Create edges (from, to) as (from.from, to).
  // Sets PathNode.to = GraphNode.value and PathNode.key.first to predecessor labels.
  nvector::iterator node;
  for(node = this->nodes.begin(); node != this->nodes.end(); ++node)
  {
    (*node).key.first = 0;
    pair_type ge_range = this->parent.getEdges((*node).from, false);
    for(usint i = ge_range.first; i <= ge_range.second; i++)
    {
      uint gnode = this->parent.edges[i].from;
      this->edges.push_back(PathEdge(gnode, node - this->nodes.begin()));
      (*node).key.first |= ((usint)1) << GCSA::ALPHABET.find(this->parent.nodes[gnode].label);
    }
    (*node).to = this->parent.nodes[(*node).from].value;
  }
  this->edge_count = this->edges.size();

  // Replace (from.from, to) with (from, to).
  // Sets PathNode.key.second to outdegree.
  this->sortEdges();
  node = this->nodes.begin();
  evector::iterator edge = this->edges.begin();
  while(node != this->nodes.end() && edge != this->edges.end())
  {
    if((*edge).from == (*node).from)
    { (*edge).from = node - this->nodes.begin(); ++edge;
      (*node).key.second++;
    }
    else { ++node; }
  }

  this->status = ready;
  return true;
}

//--------------------------------------------------------------------------

void
PathGraph::createPathNode(uint from, uint to)
{
  PathNode pn;
  pn.from = from; pn.to = to;
  pn.key = pair_type(this->parent.nodes[from].label, this->parent.nodes[to].label);
  this->nodes.push_back(pn);
}

void
PathGraph::createPathNode(const PathNode& left, const PathNode& right)
{
  PathNode pn;
  pn.from = left.from; pn.to = right.to;
  pn.key = pair_type(left.key.first, right.key.first);
  this->nodes.push_back(pn);
}

void
PathGraph::sort()
{
  this->sortByKey();

  // Update ranks.
  usint rank = 0;
  pair_type key = this->nodes[0].key;
  for(nvector::iterator iter = this->nodes.begin(); iter != this->nodes.end(); ++iter)
  {
    if((*iter).key != key) { rank++; key = (*iter).key; }
    (*iter).key = pair_type(rank, 0);
  }

  // Merge equivalent nodes.
  usint top = 0;
  pair_type node_range = EMPTY_PAIR;
  while(true)
  {
    node_range = this->nextMaximalSet(node_range);
    if(isEmpty(node_range)) { break; }
    this->nodes[top] = this->nodes[node_range.first]; top++;
  }
  this->node_count = top; this->nodes.resize(this->node_count);

  // Check if sorted.
  bool is_sorted = true;
  PathNode* candidate = &(this->nodes[0]);
  key = candidate->key; this->ranks = 1;
  for(usint i = 1; i < this->node_count; i++)
  {
    if(this->nodes[i].key != key)
    {
      if(candidate != 0) { candidate->setSorted(); }
      candidate = &(this->nodes[i]);
      key = candidate->key; this->ranks++;
    }
    else { candidate = 0; is_sorted = false; }
  }
  if(candidate != 0) { candidate->setSorted(); }
  if(is_sorted) { this->status = sorted; }
}

// Returns the next maximal mergeable set of PathNodes.
// A set of PathNodes sharing adjacent keys is mergeable, if each of the
// PathNodes begins in the same GraphNode, and no other PathNode shares
// the key. If the maximal set is empty, returns the next PathNode.
pair_type
PathGraph::nextMaximalSet(pair_type range)
{
  if(range.second >= this->node_count - 1) { return EMPTY_PAIR; }

  if(isEmpty(range)) { range = pair_type(0, 0); }
  else
  {
    range.first = range.second + 1; range.second = range.first;
    if(this->nodes[range.first - 1].key == this->nodes[range.first].key) { return range; }
  }

  for(usint i = range.second + 1; i < this->node_count; i++)
  {
    if(this->nodes[i - 1].key != this->nodes[i].key) { range.second = i - 1; }
    if(this->nodes[i].from != this->nodes[range.first].from) { return range; }
  }

  range.second = this->node_count - 1;
  return range;
}

//--------------------------------------------------------------------------

struct PathNodeComparator
{
  bool operator() (const PathNode& a, const PathNode& b) const
  {
    return (a.key < b.key);
  }
} pn_comparator;

void
PathGraph::sortByKey()
{
  parallelSort(this->nodes.begin(), this->nodes.end(), pn_comparator);
}

struct PathEdgeComparator // FIXME Cannot sort edges of multiple graphs concurrently.
{
  GraphNode* gnodes;

  // We assume that PathNodes have already been sorted by keys.
  bool operator() (const PathEdge& a, const PathEdge& b) const
  {
    return ((gnodes[a.from].label < gnodes[b.from].label) ||
            ((gnodes[a.from].label == gnodes[b.from].label) && (a.to < b.to)));
  }
} pe_comparator;

void
PathGraph::sortEdges()
{
  pe_comparator.gnodes = this->parent.nodes;
  parallelSort(this->edges.begin(), this->edges.end(), pe_comparator);
}

struct PathNodeFromComparator
{
  bool operator() (const PathNode& a, const PathNode& b) const
  {
    return (a.from < b.from);
  }
} pn_from_comparator;

void
PathGraph::sortByFrom()
{
  parallelSort(this->nodes.begin(), this->nodes.end(), pn_from_comparator);
  this->status = error;
}

struct PathEdgeFromComparator
{
  bool operator() (const PathEdge& a, const PathEdge& b) const
  {
    return (a.from < b.from);
  }
} pe_from_comparator;

void
PathGraph::sortEdgesByFrom()
{
  parallelSort(this->edges.begin(), this->edges.end(), pe_from_comparator);
}

pair_type
PathGraph::getNodesFrom(uint node)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = this->node_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(node == this->nodes[mid].from)     { high = mid; }
    else if(node < this->nodes[mid].from) { high = mid - 1; }
    else                                  { low = mid + 1; }
  }
  if(node == this->nodes[low].from) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = this->node_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(node == this->nodes[mid].from) { low = mid; }
    else                              { high = mid - 1; }
  }
  range.second = high;

  return range;
}

pair_type
PathGraph::getEdgesFrom(usint node)
{
  pair_type range;

  // Lower bound for the range.
  sint low = 0, high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low) / 2;
    if(node == this->edges[mid].from)     { high = mid; }
    else if(node < this->edges[mid].from) { high = mid - 1; }
    else                                  { low = mid + 1; }
  }
  if(node == this->edges[low].from) { range.first = low; }
  else { return EMPTY_PAIR; }

  // Upper bound for the range.
  high = this->edge_count - 1;
  while(low < high)
  {
    sint mid = low + (high - low + 1) / 2;
    if(node == this->edges[mid].from) { low = mid; }
    else                              { high = mid - 1; }
  }
  range.second = high;

  return range;
}

//--------------------------------------------------------------------------

std::vector<pair_type>*
PathGraph::getSamples(usint sample_rate)
{
  if(this->status != ready) { return 0; }

  std::vector<pair_type>* pairs = new std::vector<pair_type>();

  WriteBuffer found_nodes(this->node_count, 1);
  ReadBuffer* found = found_nodes.getReadBuffer();
  WriteBuffer sampled_nodes(this->node_count, 1);
  ReadBuffer* sampled = sampled_nodes.getReadBuffer();

  this->sortEdgesByFrom();
  std::stack<usint> active;
  active.push(this->node_count - 1); // Initial node.
  usint unsampled = 0;
  usint current = 0;
  while(!active.empty())
  {
    current = active.top(); active.pop();
    unsampled = 1;

    while(true)
    {
      if(found->readItem(current))
      {
        // Sample if:
        //   iv)  node has multiple predecessors
        if(!sampled->readItem(current))
        {
          pairs->push_back(pair_type(current, this->nodes[current].value()));
          sampled_nodes.goToItem(current); sampled_nodes.writeItem(1);
        }
        break;
      }
      found_nodes.goToItem(current); found_nodes.writeItem(1);
      pair_type range = this->getEdgesFrom(current);

      // Sample if:
      //   i)   there are no samples within sample_rate steps;
      //   ii)  current node does not have exactly one outgoing edge; or
      //   iii) (value of successor) is not (value of current node + 1).
      if(unsampled >= sample_rate || length(range) != 1 ||
         this->nodes[this->edges[range.first].to].value() != this->nodes[current].value() + 1)
      {
        pairs->push_back(pair_type(current, this->nodes[current].value()));
        sampled_nodes.goToItem(current); sampled_nodes.writeItem(1);
        for(usint i = range.first; i <= range.second; i++)
        {
          active.push(this->edges[i].to);
        }
        break;
      }

      current = this->edges[range.first].to;
      unsampled++;
    }
  }

  parallelSort(pairs->begin(), pairs->end());

  delete found;
  delete sampled;
  return pairs;
}

//--------------------------------------------------------------------------

} // namespace CSA
